@@ -1,8 +1,8 @@
 package XFST::Lookup;
 
-# $Header: /var/cvsroot/XFST-Lookup/lib/XFST/Lookup.pm,v 1.17 2005/06/27 16:24:12 alex Exp $
+# $Header: /var/cvsroot/XFST-Lookup/lib/XFST/Lookup.pm,v 1.24 2005/07/02 10:29:01 alex Exp $
 
-$VERSION="0.7";
+$VERSION="0.8";
 
 use strict;
 use Carp;
@@ -62,23 +62,40 @@ sub new {
 sub _check4bin {
     my $self = shift;
     my $binary = shift;
-    my $path2bin;
+    my $path2bin = "";
+	my $found = 0;
     
     print "Looking for '$binary'... " if $self->{debug};
 
     # try to locate xfst's lookup (or any other binary) (in $ENV{PATH})
-    # TODO: what about win32? steffi?
     if ( $^O =~ /linux|bsd|darwin/ ) {
         # got a user-defined path? add to $ENV{PATH}
         $ENV{PATH} .= ":$self->{path}" if $self->{path};
         
         # give 'which' a try...
         chomp( $path2bin = `which $binary 2>/dev/null` );
-        
-        croak "Cannot locate '$binary' in \$PATH ('$ENV{PATH}').\n" .
-            "\nYou can add a user-defined path via the 'path'-parameter.\n"
-            if ( $? != 0 );
+    
+		( $? == 0 ) ? $found = 0 : $found = 1;
     }
+	elsif ( $^O eq "MSWin32" ) {
+		my $ms_binary="$binary\.exe";
+		if( system("$ms_binary -v") == 0) {
+			print "MSWin32: $ms_binary found in standard-path...\n"
+				if $self->{debug};
+			$found=1;
+			$path2bin=$ms_binary;
+		}
+		elsif( $self->{path}  && system("$self->{path}\\$ms_binary -v") == 0 ) {
+			print "MSWin32: $ms_binary found in user-defined path: " .
+				"'$self->{path}'\n"	if $self->{debug};
+			$found = 1;
+			$path2bin="$self->{path}\\$ms_binary";
+		}
+	}
+	
+    croak "Cannot locate '$binary' in \$PATH ('$ENV{PATH}').\n" .
+    	"\nYou can add a user-defined path via the 'path'-parameter.\n"
+        unless( $found );
     
     print "$path2bin\n" if $self->{debug};
 
@@ -95,8 +112,14 @@ sub _check_lookup {
 
     my $version_str = `$bin -v 2>&1`;
     $version_str =~ /lookup ([\d\.]+)/;
-    (defined $1) ? $self->{lookup_version} = $1 :
-        carp "Cannot determine lookup-version!\n";
+	
+    if (defined $1) {
+		$self->{lookup_version} = $1;
+	}
+	else {
+		carp "Cannot determine lookup-version!\n";
+		$self->{lookup_version} = ">unknown<";
+	}
 
     # do version specific stuff here if necessary...
 
@@ -110,7 +133,7 @@ sub _check_lookup {
     
 You may provide parameters via an anonymous hashref like this:
 
-    my $test = new XFST::Lookup( {
+    my $lookup = new XFST::Lookup( {
             debug => 1 ,
             flags => 'some flags',
             path => '/home/alex/bin/:/home/roni/bin/',
@@ -139,8 +162,12 @@ you won't have to set it every time you use lookup()
 
 =item I<path> [path]
 
-Specify an extra-$PATH or a list of $PATHes in order to help XFST::Lookup
-to locate the lookup-binary. (separator: ':')
+I<Linux/*BSD/MacOS>:
+
+=over 2
+
+Specify a single addition or a list of additions to $PATH in order to help
+XFST::Lookup to locate the lookup-binary. (separator: ':')
 
 Per default XFST::Lookup is looking for the binary in $ENV{PATH} on
 Linux,*BSD and Mac OS calling the which(1)-utility that is shipped with
@@ -148,17 +175,30 @@ every *NIX-distribution.
 
 =back
 
+I<Win32>:
+
+=over 2
+
+On MSWin32 XFST::Lookup tries to locate the binary in the defaut path or
+searches a user-defined path afterwards (i.e. supplied via
+"path=>d:\alex\xerox-windows\")
+
+Manually setting the path on this OS is recommended...
+
+=back
+
 =head1 METHODS
 
 =head2 B<set_flags()> [@array|$scalar]
 
 =over 2
 
-set_flags() is used to set the flags that'll be used by *XEROX*'s
+set_flags() is used to set the flags that will be used by *XEROX*'s
 lookup-utility. See [3] for a brief overview.
 
-Please note: the 'LxL' and 'TxT'-flags are stripped, even if you set them
-manually in order to do what you mean (I at least hope so) ;)
+Please note: even if you set them manually, the 'LxL' and 'TxT'-flags are
+stripped, in order to do what you mean. XFST::Lookup needs the default
+separators in order to correctly process lookup's output.
 
 =back
 
@@ -178,25 +218,63 @@ sub set_flags {
     $self->{lookup_flags} = $flags;
 }
 
+=back
+
 =head2 B<lookup()> [{anonymous hashref}]
 
 =over 2
 
 ... is the most important method of XFST::Lookup (like the name may already
-suggest). Every time lookup() is invoked from the inside of a PERL-program,
+suggests). Every time lookup() is invoked from the inside of a PERL-program,
 *XEROX*'s lookup-utility is called with the given options and all results
 are returned as a I<hash of arrays>. This datastructure has been chosen
-because there may be ambiguous wordforms so a simple hash won't fullfil
+because there may be ambiguous wordforms so a simple hash won't fulfill
 the needs.
 
+Here's an example of the mentioned datastructure (Data::Dumper-format):
+
+    $VAR1 = 'eine';
+    $VAR2 = [
+              'ein+ART+NOM+SG',
+              'ein+ART+AKK+SG',
+              'ein+PIS+AKK+SG',
+              'ein+PIS+NOM+SG',
+              'ein+PIAT+AKK+SG',
+              'ein+PIAT+NOM+SG',
+              'eine+PIDAT'
+            ];
+    $VAR3 = 'Haus';
+    $VAR4 = [
+              'Haus+NN+NOM+SG',
+              'Haus+NN+DAT+SG',
+              'Haus+NN+AKK+SG'
+            ];
+    $VAR5 = 'das';
+    $VAR6 = [
+              'das+ART+NOM+SG',
+              'das+ART+AKK+SG',
+              'das+KOUS',
+              'der+PDAT+NOM+SG',
+              'der+PDAT+AKK+SG',
+              'der+PDS+NOM+SG',
+              'der+PDS+AKK+SG'
+            ];
+    ...
+
+While the hash's key contains a I<wordform>, the value provides an @array
+of all analyzed I<readings> of this wordform.
+
+The following parameters are available to alter XFST::Lookup's runtime
+behaviour:
+
 =over 2
 
 =item I<fst> [path]
 
-set the patch to the file that contains the finite state transducer to use
+set the path to the file that contains the finite state transducer to use
 ('save stack'-format).
 
-B<fst> is optional at this point iff it was already passed to the constructor,
+I<fst> is optional at this point iff it was already passed to the constructor,
 it is forbidden if you use the I<strategy>-parameter at the same time.
 
 Please note: Setting B<fst> again at this point will overwrite the setting for
@@ -228,7 +306,7 @@ sub lookup {
     
     my %results;
 
-    my %seen;   # remove until next version
+    my %seen;
     
     croak "Options not supplied via hashref!\n" unless ( ref($opt) eq 'HASH' );
     
@@ -247,19 +325,19 @@ sub lookup {
     if ( $opt->{fst} ) {
         $lookup_file= $opt->{fst};
         print STDERR "Using a single FST ('$lookup_file').\n" if $self->{debug};
+    	croak "File '$lookup_file' doesn't exist!\n" unless -e $lookup_file;
     }
     else {
+    	croak "File '$lookup_file' doesn't exist!\n" unless -e $opt->{strategy};
         $lookup_file= "-f " . $opt->{strategy};
         print STDERR "Using a lookup-strategy ('$lookup_file').\n"
             if $self->{debug};
     }
-   
-    croak "File '$lookup_file' doesn't exist!\n" unless -e $lookup_file;
 
     print "Using FST '$lookup_file'...\n", scalar(@{$opt->{words}}),
         " words to look up.\n" if $self->{debug};
 
-    # any extra flags that shall be passed to *XEROX*'s lookup?
+    # Any extra flags that shall be passed to *XEROX*'s lookup?
     my $flags;
     ( $self->{lookup_flags} ) ? $flags = '' :
         $flags="-flags $self->{lookup_flags}";
@@ -271,52 +349,89 @@ sub lookup {
         print STDERR "Stripping flags... (new: '$flags')\n" if $self->{debug};
     }
 
-    # where shall we direct lookup's STDERR output to?
-    # we don't want to parse that one anyway... (infos, stats)
-    # TODO: what about win32? :(
+    # Where shall we direct lookup's STDERR output to?
+    # We don't want to parse that one anyway... (infos, stats)
+    # NOTE: Doesn't work on MS Windows...
     my $dest_stream;
     ( $self->{debug} == 1 ) ? $dest_stream = '&2' :  
         ( ( $^O =~ /linux|bsd|darwin/i ) ?
             $dest_stream = '&2' : $dest_stream = '/dev/null' );
     
-    # using fork() and stuff in order to get rid of File::Temp...
+    # Using fork() and stuff in order to get rid of File::Temp...
+	# ...iff not running on MS Windows...
+	
+    my ( $pid, @output );
 
-    my ( $pid, @output ); 
-    # coming up next: fork()-magic ;-)
-    if( $pid = open( CHILD, "-|")) { # -> parent
-        print STDERR "Child forked. PID is $pid.\n" if $self->{debug};
+	if ( $^O eq "MSWin32" ) {
+		# This is a hack - I am aware of that :(
+		# I don't know how to do it any better on Win32, because I am not
+        # familiar with that OS at all.
 
-        chomp( @output = <CHILD> );
+		# Here's what went wrong/didn't work/caught fire:
+		#  o Fcntl (flock) doesn't work at all (see perlport-function)
+		#  o wait/waitpid doesn't work as expected (doesn't seem to threat
+		#    a pipe as a child-process - but it should (according to
+        #    perlport-function again), because perl is able to emulate fork()
+        #    although win32 generally lacks support for this syscall.
+		#  o IPC::Open2 didn't work well - in fact it refused to let me read
+		# 	 any data back from the pipe
+		# 	 
+		# 	 ...so File::Temp and a pseudo-cat are the last resort :(		
+		
+		print "Using File::Temp on Win32 :(\n" if $self->{debug};
 
-        close CHILD;
-    }
-    else { # -> child
-        croak "fork() failed: '$!'\n" unless defined $pid;
-        print STDERR "Child executes $self->{lookup_binary}." .
-            " Output goes to parent.\n" if $self->{debug};
+		use File::Temp;
+		
+  		my $tmp = new File::Temp( UNLINK => 1, SUFFIX => '.txt' );
 
-        # open the pipe to lookup...
-        open PIPE,
-            "| $self->{lookup_binary} $lookup_file $flags 2>$dest_stream"
-                or croak "Cannot fork(): $self->{lookup_binary}: '$!'\n";
-        # ...and feed in the wordforms.
-        print PIPE join("\n",@{$opt->{words}}), "\n";
+		print "Tempfile is '$tmp'\n" if $self->{debug};
+
+		# write all wordforms to the tempfile...
+		print $tmp join("\n", @{$opt->{words}}), "\n";
+		
+		# ... use PERL as kind of cat-command to pipe the wordforms to
+		# lookup. This looks terribly ugly, but the Win32 echo-command
+		# doesn't even seem to handle escape-sequences, which are required
+		# in order to get some useful results out of lookup.
+		open WINPIPE, "perl -e \"open FH, $tmp or die; print foreach <FH>\"" . 
+			" | $self->{lookup_binary} $lookup_file $flags |" or
+			die "Failed to open pipes: '$!'\n";
+
+		@output = <WINPIPE>;
+	}	
+	else {	
+        # running on *NIX -- Using fork()'s magic ;-)
+        if( $pid = open( CHILD, "-|") ) { # -> parent
+            print STDERR "Child forked. PID is $pid.\n" if $self->{debug};
+
+            chomp( @output = <CHILD> );
+
+            close CHILD;
+        }
+        else { # -> child
+            croak "fork() failed: '$!'\n" unless defined $pid;
+            print STDERR "Child executes $self->{lookup_binary}." .
+                " Output goes to parent.\n" if $self->{debug};
+
+            # open the pipe to lookup...
+            open PIPE,
+                "| $self->{lookup_binary} $lookup_file $flags 2>$dest_stream"
+                    or croak "Cannot fork(): $self->{lookup_binary}: '$!'\n";
+            # ...and feed in the wordforms.
+            print PIPE join("\n",@{$opt->{words}}), "\n";
         
-        # terminate child (continuing under parent-pid)
-        exit;
-    }
+            # terminate child (continuing under parent-pid)
+            exit;
+        }
+	}
 
-    # the follwing class describes all chars that are valid for the
-    # transducer's output.
-    my $cclass = '\w\d\.-_\+!\"\'\\\//';
-    
     # extract information from lookup's output
     foreach my $line ( @output ) {
         chomp $line;
         next unless $line;
 
         # match the three fields ( 1x input, 2x parts analysis(base+tags) )
-        $line =~ /^([$cclass]+)\s+([$cclass]+)\s+([$cclass]+)$/;
+        $line =~ /^([^\s]+)\s+([^\s]+)\s+([^\s]+)$/;
         if ( $1 && $2 && $3 ) {
             unless( $seen{$1}->{"$2$3"} ) {
                 push @{$results{$1}}, "$2$3";
@@ -327,7 +442,7 @@ sub lookup {
             carp "Error parsing line: '$line'\n";
         }
     }
-    
+
     return %results;
 }
 
